@@ -24,14 +24,15 @@ from azureml.core import Environment, Experiment, Run, RunConfiguration, ScriptR
 from azureml.core.runconfig import DockerConfiguration, MpiConfiguration
 from azureml.data import OutputFileDatasetConfig
 from azureml.data.dataset_consumption_config import DatasetConsumptionConfig
-from azureml.train.hyperdrive import HyperDriveConfig
+from azureml.train.hyperdrive import HyperDriveConfig, GridParameterSampling, PrimaryMetricGoal, choice
+from azureml.dataprep.fuse.daemon import MountContext
 
 from health_azure.utils import (create_python_environment, create_run_recovery_id, _find_file,
                                 is_run_and_child_runs_completed, is_running_in_azure_ml, register_environment,
                                 run_duration_string_to_seconds, to_azure_friendly_string, RUN_CONTEXT, get_workspace,
-                                PathOrString)
+                                PathOrString, DEFAULT_ENVIRONMENT_VARIABLES)
 from health_azure.datasets import (DatasetConfig, StrOrDatasetConfig, _input_dataset_key, _output_dataset_key,
-                                   _replace_string_datasets)
+                                   _replace_string_datasets, setup_local_datasets)
 
 logger = logging.getLogger('health_azure')
 logger.setLevel(logging.DEBUG)
@@ -63,6 +64,10 @@ class AzureRunInfo:
     """A list of folders that contain all the datasets that the script uses as outputs. Output datasets must be
          specified when calling `submit_to_azure_if_needed`. Here, they are made available as Path objects. If no output
          datasets are specified, the list is empty."""
+    mount_contexts: List[MountContext]
+    """A list of mount contexts for input datasets when running outside AzureML. There will be a mount context
+    for each input dataset where there is no local_folder, there is a workspace, and use_mounting is set.
+    This list is maintained only to prevent exit from these contexts until the RunInfo object is deleted."""
     run: Optional[Run]
     """An AzureML Run object if the present script is executing inside AzureML, or None if outside of AzureML.
     The Run object has methods to log metrics, upload files, etc."""
@@ -131,7 +136,8 @@ def create_run_configuration(workspace: Workspace,
             private_pip_wheel_path=private_pip_wheel_path,
             docker_base_image=docker_base_image,
             environment_variables=environment_variables)
-        register_environment(workspace, run_config.environment)
+        registered_env = register_environment(workspace, run_config.environment)
+        run_config.environment = registered_env
     else:
         raise ValueError("One of the two arguments 'aml_environment_name' or 'conda_environment_file' must be given.")
 
@@ -147,6 +153,8 @@ def create_run_configuration(workspace: Workspace,
     if max_run_duration:
         run_config.max_run_duration_seconds = run_duration_string_to_seconds(max_run_duration)
 
+    # Create MPI configuration for distributed jobs (unless num_splits > 1, in which case
+    # an AML HyperdriveConfig is instantiated instead
     if num_nodes > 1:
         distributed_job_config = MpiConfiguration(node_count=num_nodes)
         run_config.mpi = distributed_job_config
@@ -164,6 +172,35 @@ def create_run_configuration(workspace: Workspace,
     return run_config
 
 
+def create_crossval_hyperdrive_config(num_splits: int,
+                                      cross_val_index_arg_name: str = "cross_validation_split_index",
+                                      metric_name: str = "val/loss") -> HyperDriveConfig:
+    """
+    Creates an Azure ML HyperDriveConfig object for running cross validation. Note: this config expects a metric
+    named <metric_name> to be logged in your training script([see here](
+    https://docs.microsoft.com/en-us/azure/machine-learning/how-to-tune-hyperparameters#log-metrics-for-hyperparameter-tuning))
+
+    :param num_splits: The number of splits for k-fold cross validation
+    :param cross_val_index_arg_name: The name of the argument received by each of the child runs that indicates which
+        split that child represents.
+    :param metric_name: The name of the metric that the HyperDriveConfig will compare runs by. Please note that it is
+        your responsibility to make sure a metric with this name is logged to the Run in your training script
+    :return: an Azure ML HyperDriveConfig object
+    """
+    logging.info(f"Creating a HyperDriveConfig. Please be aware that this expects to find the metric {metric_name}"
+                 f" logged to the Run during your training script.")
+    return HyperDriveConfig(
+        run_config=ScriptRunConfig(""),
+        hyperparameter_sampling=GridParameterSampling(
+            {
+                cross_val_index_arg_name: choice(list(range(num_splits)))
+            }),
+        primary_metric_name=metric_name,
+        primary_metric_goal=PrimaryMetricGoal.MINIMIZE,
+        max_total_runs=num_splits
+    )
+
+
 def create_script_run(snapshot_root_directory: Optional[Path] = None,
                       entry_script: Optional[PathOrString] = None,
                       script_params: Optional[List[str]] = None) -> ScriptRunConfig:
@@ -295,7 +332,8 @@ def submit_to_azure_if_needed(  # type: ignore
         submit_to_azureml: Optional[bool] = None,
         tags: Optional[Dict[str, str]] = None,
         after_submission: Optional[Callable[[Run], None]] = None,
-        hyperdrive_config: Optional[HyperDriveConfig] = None) -> AzureRunInfo:  # pragma: no cover
+        hyperdrive_config: Optional[HyperDriveConfig] = None
+) -> AzureRunInfo:  # pragma: no cover
     """
     Submit a folder to Azure, if needed and run it.
     Use the commandline flag --azureml to submit to AzureML, and leave it out to run locally.
@@ -370,9 +408,13 @@ def submit_to_azure_if_needed(  # type: ignore
         submit_to_azureml = AZUREML_COMMANDLINE_FLAG in sys.argv[1:]
     if not submit_to_azureml:
         # Set the environment variables for local execution.
-        if environment_variables is not None:
-            for k, v in environment_variables.items():
-                os.environ[k] = v
+        environment_variables = {
+            **DEFAULT_ENVIRONMENT_VARIABLES,
+            **(environment_variables or {})
+        }
+
+        for k, v in environment_variables.items():
+            os.environ[k] = v
 
         output_folder = Path.cwd() / OUTPUT_FOLDER
         output_folder.mkdir(exist_ok=True)
@@ -380,9 +422,14 @@ def submit_to_azure_if_needed(  # type: ignore
         logs_folder = Path.cwd() / LOGS_FOLDER
         logs_folder.mkdir(exist_ok=True)
 
+        mounted_input_datasets, mount_contexts = setup_local_datasets(aml_workspace,
+                                                                      workspace_config_path,
+                                                                      cleaned_input_datasets)
+
         return AzureRunInfo(
-            input_datasets=[d.local_folder for d in cleaned_input_datasets],
+            input_datasets=mounted_input_datasets,
             output_datasets=[d.local_folder for d in cleaned_output_datasets],
+            mount_contexts=mount_contexts,
             run=None,
             is_running_in_azure_ml=False,
             output_folder=output_folder,
@@ -413,12 +460,13 @@ def submit_to_azure_if_needed(  # type: ignore
         num_nodes=num_nodes,
         max_run_duration=max_run_duration,
         input_datasets=cleaned_input_datasets,
-        output_datasets=cleaned_output_datasets,
+        output_datasets=cleaned_output_datasets
     )
     script_run_config = create_script_run(snapshot_root_directory=snapshot_root_directory,
                                           entry_script=entry_script,
                                           script_params=script_params)
     script_run_config.run_config = run_config
+
     if hyperdrive_config:
         config_to_submit: Union[ScriptRunConfig, HyperDriveConfig] = hyperdrive_config
         config_to_submit._run_config = script_run_config
@@ -513,6 +561,7 @@ def _generate_azure_datasets(
     return AzureRunInfo(
         input_datasets=returned_input_datasets,  # type: ignore
         output_datasets=returned_output_datasets,  # type: ignore
+        mount_contexts=[],
         run=RUN_CONTEXT,
         is_running_in_azure_ml=True,
         output_folder=Path.cwd() / OUTPUT_FOLDER,
@@ -526,7 +575,7 @@ def append_to_amlignore(lines_to_append: List[str], amlignore: Optional[Path] =
     the context.
     If the file does not exist yet, it will be created, the contents written, and deleted when leaving the context.
 
-    :param lines_to_append: The text lines that should be added at the end of the .amlignore file
+    :param lines_to_append: The text lines that should be added at the enund of the .amlignore file
     :param amlignore: The path of the .amlignore file that should be modified. If not given, the function
         looks for a file in the current working directory.
     """
@@ -585,7 +634,8 @@ def main() -> None:
         compute_cluster_name=args.compute_cluster_name,
         snapshot_root_directory=Path(args.snapshot_root_directory),
         entry_script=Path(args.entry_script),
-        conda_environment_file=Path(args.conda_environment_file))
+        conda_environment_file=Path(args.conda_environment_file),
+    )
 
 
 if __name__ == "__main__":
